from typing import Callable, Any, Optional, List
from pymonad.tools import curry
from collections import namedtuple
from metis_fn import monad, fn

from . import pip, app

activity_token = ":"

pdp_value = namedtuple("PdpValue", "pip, name, ctx, namespace, activities, decisions")

#
# PDPs
#
def token_pdp(pdp_value):
    """
    Interrogates the PIP which holds the id_token to determine it is valid.  The ID Token has been inserted in
    the PIP by the app being configured to use the subject_token module.

    The id_token in the PIP MUST be present and it MUST be valid according to the logic in the fn crypto_util.decode_jwt.
    """
    pdp_value.decisions.append(pdp_value.pip.token_valid())
    return pdp_value

def activity_pdp(pdp_value) -> monad.MEither:
    """
    Takes Activities from an activity collection, and tests the provided context against them.
    Example:
        $ activity_policy(['service:resource:action:action1'])("service:resource:action:action2")
    """
    if pdp_value.activities is None:
        return False
    pdp_value.decisions.append(any(activity_matcher(pdp_value.ctx, activity) for activity in
                                   filter_for_service(pdp_value.namespace, pdp_value.activities)))
    return pdp_value


#
# Decorators
#
def pdp_decorator(name: str,
                  pdps: List[callable],
                  ctx: str = None,
                  namespace: Optional[str] = None,
                  error_cls: Any = None):
    """
    A general PDP decorator.  It can run any PDP, the callables are provided in the pdps list.
    """

    def inner(f):
        def invoke(*args, **kwargs):
            pip = _get_pip(args, kwargs)
            if pip is None:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

            results = fn.compose_iter(pdps, pdp_value(pip, name, ctx, namespace, pip.subject.value.activities(), []))

            if all(results.decisions):
                return f(*args, **kwargs)
            else:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

        return invoke

    return inner


def token_pdp_decorator(name: str,
                        ctx: str = None,
                        namespace: Optional[str] = None,
                        error_cls: Any = None,
                        pdp: callable = token_pdp):
    """
    Policy to check the presence and validity of a bearer token.

    The decorated function must have a kwarg called "pip" which is an instance of pip.PipConfig.
    + The pip has fn "id_token", which returns a monad whose value is a token; id_token: monad.EitherMonad[crypto_util.IdToken]
    """

    def inner(fn):
        def invoke(*args, **kwargs):
            pip = _get_pip(args, kwargs)
            if pip is None or pip.id_token is None:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

            result = pdp(pdp_value(pip, name, ctx, namespace, pip.subject_activities(), []))

            if all(result.decisions):
                return fn(*args, **kwargs)
            else:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

        return invoke

    return inner

def _get_pip(args, kwargs):
    if args and isinstance(args[0], pip.Pip):
        return args[0]
    if (pip_from_kw:=kwargs.get('pip', None)):
        return pip_from_kw
    if args and isinstance(args[0], app.Request):
        return args[0].pip
    if (req_from_kw:=kwargs.get('request', None)):
        return req_from_kw.pip
    return None


def activity_pdp_decorator(name: str,
                           ctx: str,
                           namespace: Optional[str] = None,
                           error_cls: Any = None,
                           pdp: callable = activity_pdp):
    """
    Adds activity-based policy checking to the function.

    + name: (str); to be included in the error_cls on error.
    + ctx:  (str); the activity being tested against.
    + namespace (Optional[str])
    + error_cls: (Optional[Any]); In the case of an authorisation error, the error class to be used to encapsulate the failure.
                 If not provides, a monad wrapped str is returned.

    The decorated function must have a kwarg called "pip" which is an instance of pip.PipConfig.
    + The pip has fn "subject", which returns a monad whose value has an activities fn which returns List[str].
      pip.PipConfig implements this protocol
      e.g pip.subject.value.activities => ['activity1', 'activity2']
    """

    def inner(fn):
        def invoke(*args, **kwargs):
            pip = kwargs.get('pip', None)
            if pip is None:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

            result = pdp(pdp_value(pip=pip,
                                   name=name,
                                   ctx=ctx,
                                   namespace=namespace,
                                   activities=pip.subject_activities(),
                                   decisions=[]))


            if all(result.decisions):
                return fn(*args, **kwargs)
            else:
                return monad.Left(error_cls(message="Unauthorised",
                                            name=name,
                                            code=401)) if error_cls else monad.Left("Unauthorised")

        return invoke

    return inner


#
# Helpers
#

def filter_for_service(namespace, activities: list) -> list:
    if not namespace:
        return activities
    return list(filter(lambda ct: namespace in ct, activities))


@curry(2)
def activity_matcher(context, activity):
    return matcher(fn.rest(context.split(activity_token)), fn.rest(activity.split(activity_token)))


def matcher(ctx_xs: list, activity_xs: list):
    ctx_fst, ctx_rst = fn.first(ctx_xs), fn.rest(ctx_xs)
    act_fst, act_rst = fn.first(activity_xs), fn.rest(activity_xs)
    if ctx_fst is None:
        return True
    if not ismatch(ctx_fst, act_fst):
        return False
    return matcher(ctx_rst, act_rst)


def ismatch(ctx_token, act_token):
    return act_token == "*" or ctx_token == act_token



from typing import Dict, Tuple
from pymonad.tools import curry
import requests

from metis_fn import monad
from . import error

class HttpError(error.BaseError):
    pass

class BearerTokenAuth(requests.auth.AuthBase):
    """
    Custom Bearer token Authn
    """
    def __init__(self, token):
        self.token = token
    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


def extract_fn_raw(response):
    return response.text

def extract_fn_json(response):
    return response.json()

def extract_fn_text(response):
    return response.text()

def extract_by_content_type(response):
    """
    Looks for the content type in the response; e.g. application/json; charset=utf-8
    and find the appropriate extract fn and apply the response to that fn.  Otherwise return raw.
    """
    factory = {'application/json': extract_fn_json, 'application/text': extract_fn_text}
    return factory.get(response.headers['Content-Type'].split(";")[0], extract_fn_raw)(response)

@curry(3)
def http_response_monad(step, extract_fn, response) -> Tuple[int, dict]:
    """
    Only 200-series http status codes are successes.  400 and 500 series are failures.
    It is assumed that a REST call to a service will not return a 300-series code.
    """
    if response.value.status_code in [200,201]:
        return monad.Right((response.value.status_code, extract_fn(response.value)))
    else:
        return monad.Left(HttpError(message=format_error(response.value),
                                    name=step,
                                    code=response.value.status_code,
                                    ctx=extract_fn(response.value),
                                    retryable=http_retryable_status(response.value.status_code)))

def http_retryable_status(code):
    return code >= 500

def format_error(resp: requests.models.Response) -> str:
    return "HTTP Error Response; {mth} ; {url} ; {reason}".format(mth=resp.request.method,
                                                                  url=resp.request.url,
                                                                  reason=resp.reason)


from typing import Any, Callable, Union, Optional, Protocol, TypeVar, Type
from datetime import datetime
from metis_fn import monad, chronos, singleton

from . import error, state_machine

T = TypeVar('T', bound='CircuitStateProviderProtocol')

# Circuit States, transitions, and state machine
transition_failure           = 'failure'
transition_persistent_failure = 'persistent_failure'
transition_success           = 'success'
state_half_open              = 'half_open'
state_half_closed            = 'half_closed'
state_open                   = 'open'
state_closed                 = 'closed'
circuit_state_map = state_machine.state_transition_map([(None,              transition_failure,            state_half_open),
                                                       (None,               transition_persistent_failure, state_open),
                                                       (None,               transition_success,            state_closed),
                                                       (state_half_open,    transition_persistent_failure, state_open),
                                                       (state_half_closed,  transition_persistent_failure, state_open),
                                                       (state_half_closed,  transition_failure,            state_half_open),
                                                       (state_half_open,    transition_success,            state_closed),
                                                       (state_open,         transition_success,            state_half_closed),
                                                       (state_closed,       transition_success,            state_closed),
                                                       (state_half_closed,  transition_success,            state_closed)])


class CircuitOpen(error.BaseError):
    pass


class CircuitStateProviderProtocol(Protocol):
    def __init__(self, circuit_name: Union[None, str]):
        ...

    def circuit_state(self) -> Union[None, str]:
        """
        Returns the current circuit state or None.  The circuit state machine is provided by the circuit breaker to the provider.
        It comes from the state machine defined in circuit_state_map
        """
        ...

    def failures(self) -> int:
        ...

    def last_state_chg_time(self) -> datetime:
        ...

    def update_state(self, failures: int,  last_state_chg_time: datetime, circuit_state: Union[None, str]) -> T:
        """
        Expects the circuit state to be updated
        """
        ...

class CircuitConfiguration(singleton.Singleton):
    # Factors that determine if a circuit should be placed in open state.
    # 3 failures in a 5 min period, opens the circuit
    failure_threshold_seconds = 5 * 60 # 5 minutes
    failure_count_threshold = 3

    # The number of minutes from the time a circuit transitioned to open before it can be retried
    open_stand_down_period = 5 * 60

    def configure(self,
                  max_retries: int =None,
                  circuit_state_provider: Optional[CircuitStateProviderProtocol] = None):
        self.circuit_state_provider = circuit_state_provider
        self.max_retries = 3 if max_retries is None else max_retries # used by the backoff decorator to configure the number of retry attempts
        pass

    def provider(self):
        return self.circuit_state_provider if hasattr(self, "circuit_state_provider") else None


def max_retries():
    return CircuitConfiguration().max_retries if hasattr(CircuitConfiguration(), 'max_retries') else None

def circuit_breaker():
    """
    Circuit Breaker Decorator.

    The kwargs to the wrapped function MAY include a 'circuit_state_provider' argument which can manage the state of the circuit
    over multiple invocations.  Its not implemented here, but must be injected in the args.
    If its available it must support the following methods:
    + circuit_state
    + failures
    + last_state_chg_time: Takes and returns the last state change time as a ISO8601 formatted str.
    + update_state

    When a circuit_state_provider is not provided, the circuit is a no-op.

    """
    def inner(fn):
        def breaker(*args, **kwargs):
            circuit_state_provider = get_a_provider(kwargs, CircuitConfiguration())
            if circuit_state_provider and is_open(circuit_state_provider) and is_in_stand_down_period(circuit_state_provider.last_state_chg_time):
                return monad.Left(CircuitOpen(message="Circuit Open",
                                              code=500,
                                              ctx={'circuit_state': circuit_state_provider.circuit_state, 'failures': circuit_state_provider.failures}))

            result = fn(*args, **kwargs)

            if circuit_state_provider:
                if result.is_left():
                    circuit_failure(circuit_state_provider)
                else:
                    transition_circuit_on_success(circuit_state_provider)
            return result
        return breaker
    return inner

def get_a_provider(from_args, from_config):
    """
    From Args takes precidence.
    """
    args_provider = from_args.get('circuit_state_provider', None)
    return from_config.provider() if args_provider is None else args_provider

def monad_failure_predicate(monad_result: Union[Any, monad.MEither]) -> bool:
    return isinstance(monad_result, monad.MEither) and monad_result.is_left()

def http_retryable_monad_failure_predicate(monad_result: Union[Any, monad.MEither]) -> bool:
    """
    HTTP Client-based errors (400-series) are not retryable 
    """
    if not isinstance(monad_result, monad.MEither):
        return False
    if monad_result.is_left() and isinstance(monad_result.error(), str):
        return False
    return monad_result.is_left() and monad_result.error().retryable

def circuit_failure(circuit_state_provider: Any) -> Any:
    if exhasted_failures_over_period(last_state_chg_time=circuit_state_provider.last_state_chg_time, failures=circuit_state_provider.failures):
        open_circuit(circuit_state_provider)
        pass
    else:
        update_circuit_failures(circuit_state_provider)
    pass

def transition_circuit_on_success(circuit_state_provider: Any) -> Any:
    transition = circuit_transition(from_state=circuit_state_provider.circuit_state, with_transition=transition_success)
    if circuit_state_provider.circuit_state != transition.value:
        circuit_state_provider.update_state(circuit_state=transition.value,
                                            last_state_chg_time=chronos.time_now(tz=chronos.tz_utc()),
                                            failures=0)
    return circuit_state_provider

def exhasted_failures_over_period(last_state_chg_time: datetime, failures: int) -> bool:
    if last_state_chg_time is None:
        return False
    return failures_within_time_threshold(last_state_chg_time) and (failures >= (CircuitConfiguration().failure_count_threshold) )

def failures_within_time_threshold(circuit_time: datetime) -> bool:
    """
    When the circuit is half open, we can retry until we reach the failure threshold.
    """
    return (chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()]) - circuit_time.timestamp()) < CircuitConfiguration().failure_threshold_seconds

def is_in_stand_down_period(last_state_chg_time: datetime) -> bool:
    """
    When the circuit is open there is a stand down period where the circuit will not be retried
    """
    return (chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()]) - last_state_chg_time.timestamp()) < CircuitConfiguration().open_stand_down_period

def open_circuit(circuit_state_provider: Any) -> Any:
    if circuit_state_provider.circuit_state == state_open:
        return circuit_state_provider

    circuit_state_provider.update_state(circuit_state=circuit_transition(from_state=circuit_state_provider.circuit_state, with_transition=transition_persistent_failure).value,
                                        last_state_chg_time=chronos.time_now(tz=chronos.tz_utc()),
                                        failures=0)
    return circuit_state_provider

def update_circuit_failures(circuit_state_provider: Any) -> Any:
    transition = circuit_transition(from_state=circuit_state_provider.circuit_state, with_transition=transition_failure).value
    if transition !=circuit_state_provider.circuit_state:
        circuit_state_provider.update_state(circuit_state=transition,
                                            last_state_chg_time=chronos.time_now(tz=chronos.tz_utc()),
                                            failures=1 if circuit_state_provider.failures is None else circuit_state_provider.failures + 1)

    return circuit_state_provider

#
# Circuit State Management
#
def circuit_transition(from_state: str, with_transition: str) -> monad.MEither:
    return state_machine.transition(state_map=circuit_state_map, from_state=from_state, with_transition=with_transition)

def is_open(config):
    return config.circuit_state == state_open



import json
from dataclasses import make_dataclass, dataclass
from functools import reduce
from typing import Tuple, Union, Optional
from .crypto_util import jwk, jwt
from pymonad.tools import curry

from metis_fn import monad, chronos
from . import error


class JwtDecodingError(error.BaseError):
    pass


@dataclass
class IdTokenClaims:
    iss: str
    sub: str
    aud: str
    iat: int
    exp: int
    azp: str


class IdToken:

    def __init__(self, jwt, claims=None):
        self.jwt = jwt
        self.claims = claims

    def id_token_claim_types(self):
        return IdTokenClaims.__annotations__.keys()

    def id_claims(self):
        def only_id_token_claim(acc, claim_key):
            clm = self.claims.get(claim_key, None)
            if clm:
                acc[claim_key] = clm
            return acc

        return IdTokenClaims(**reduce(only_id_token_claim, self.id_token_claim_types(), {}))

    def all_claims(self):
        cls = self._dynamic_dataclass()
        return cls(**self.claims)

    def _dynamic_dataclass(self):
        return make_dataclass("TokenClaims", [(k, type(v).__name__) for k, v in self.claims.items()])

    def sub(self):
        return self.claims['sub']

    def exp(self):
        return self.claims['exp']

    def expired(self):
        return self.exp() < int(chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()]))


def parse_generate_id_token(serialised_jwt, jwks=None, claims_to_assert=None):
    """
    Takes an encoded JWT and returns an verified id_token
    """
    return monad.Right(jwks) >> decode_jwt(claims_to_assert, serialised_jwt) >> to_id_token


def bearer_token_valid(serialised_jwt: str) -> bool:
    return parse_generate_id_token(serialised_jwt) >> validate


#
# Helpers
#

def validate(token):
    return not token.expired()


@curry(3)
@monad.monadic_try(name="decode_jwt", status=401, error_cls=JwtDecodingError)
def decode_jwt(claims_to_assert, serialised_jwt: str, jwks: Optional[jwk.JWKSet]) -> Tuple:
    """
    Parses the jwt, validing the signature (from JWKS) and the EXP/AUD claims
    """
    if jwks is None:
        return serialised_jwt, json.loads(jwt.JWT(jwt=serialised_jwt).token.objects['payload'])
    return serialised_jwt, jwt.JWT(jwt=serialised_jwt, key=jwks, check_claims=claims_to_assert)


def to_id_token(jwt_claims: Tuple[str, Union[dict, jwt.JWT]]):
    """
    Takes a Tuple of the serialised JWT and one of a dict or jwt.JWT.
    When a Dict is provided is it assumed to be the already decoded claims; this occurs when the JWT is not decoded with a JWK
    """
    serialised_jwt, decoded_jwt = jwt_claims
    if isinstance(decoded_jwt, dict):
        return monad.Right(IdToken(serialised_jwt, decoded_jwt))
    return monad.Right(IdToken(serialised_jwt, (json.loads(decoded_jwt.claims))))


def bearer_token_hdr(jwt):
    return {'authorization': "Bearer {}".format(jwt)}



from functools import reduce
from aws_lambda_powertools.utilities.data_classes import (
    S3Event,
    APIGatewayProxyEvent,
    KafkaEvent)
from aws_lambda_powertools.utilities.data_classes.s3_event import S3EventRecord

from typing import Callable, Any, Type

from metis_fn import monad, fn, chronos
from . import (env as environment,
               span_tracer,
               logger,
               app_events,
               app_value,
               app_route,
               app_serialisers, observable)

DEFAULT_SUCCESS_HTTP_CODE = 200
DEFAULT_FAILURE_HTTP_CODE = 400

"""
App provides a number of helpers to build a Lambda event handling pipeline.  

@app.route
---------- 
Maintains simple routing state; mapping a function to a route "symbol.  The function will receive a app_value.Request object 
and is expected to pass it back wrapped in an monad.MEither, with the following state options (which will be understood by the 
app.responder fn.
+ monad.Right(app_value.Request.response), with a value JSON serialisable.
+ monad.Left(app_value.Request.response), with a value JSON serialisable.
+ monad.Left(app_value.Request), with error property containing an object which responds to 'error()' and returns a value JSON serialisable.

Example: 
@app.route('s3_bucket_name_part')
def run_a_handler(request: app_value.RequestEvent) -> monad.MEither:
    return handler(request=request)

Another approach is to use the 3-part tuple form for the route template.  This is useful when wanting to pattern match on the 
event.  For an API Gateway event, the event kind is constructed like this:

                        ('API', {METHOD}, {path-template}); for example...

> ('API', 'GET', '/resourceBase/resource/uuid1')

Therefore a route defined as follows will match this pattern:
> @app.route(('API', 'GET', '/resourceBase/resource/{id}'))

Additionally, the app_value.Request.event property will (an instance of ApiGatewayapp_value.RequestEvent) will include a 'path_params'
property (dict) which will extract the templated arguments; for example:

> {'id': uuid1}     

 

app.pipeline
------------
The main event pipeline handler.  Your lambda initiates the pipeline by calling this function with:
+ the aws event.
+ the aws event context
+ the env (as a str)
+ an optional request parser
+ an optional policy information point initiator
+ a handler guard condition function.

It parses the event, using hints within the event to determine the app.route to be called.
+ S3 events.  The object S3StateChangeEvent is created with a collection of S3Object.  S3StateChangeEvent.kind is 
              used to determine the app.route symbol.  The fn domain_from_bucket_name() collects the unique bucket names
              (there should only be one), and takes the most significant part based on the default separator (DEFAULT_S3_BUCKET_SEP)
              This then is the symbol expected on an app.route.  

On Errors.  Remember when you want your handler to generate an error that is parsable by the responder:
+ Use the request.error property to hold the error.
+ Unwrap any monad before adding to request.error.
+ Use app_value.AppError as a super type for your error as it sets up a JSON serialiser and returns it on the error() call wrapping the 
  exception message generated by the error.  You can implement your own in the fashion of the serialisers in app_serialisers 
"""

DEFAULT_RESPONSE_HDRS = {'Content-Type': 'application/json'}

Request = app_value.Request
RequestEvent = app_value.RequestEvent
ApiGatewayRequestEvent = app_value.ApiGatewayRequestEvent
S3StateChangeEvent = app_value.S3StateChangeEvent
S3Object = app_value.S3Object
KafkaRecordsEvent = app_value.KafkaRecordsEvent
EventBridgePublishEvent = app_value.EventBridgePublishEvent

Serialiser = app_serialisers.SerialiserProtocol
DictToJsonSerialiser = app_serialisers.DictToJsonSerialiser
DictToJsonLDSerialiser = app_serialisers.DictToJsonLDSerialiser

AppError = app_value.AppError


def route(pattern, opts=None):
    return app_route.route(pattern, opts)


def pipeline(event: dict,
             context: dict,
             env: environment.EnvironmentProtocol,
             params_parser: Callable,
             pip_initiator: Callable,
             handler_guard_fn: Callable,
             event_source_cls: Type[S3Event | APIGatewayProxyEvent] | None = None,
             factory_overrides: dict = None):
    """
    Runs a general event handler pipeline.  Initiated by the main handler function.

    It takes the Lambda Event and Context objects.  It builds a request object and invokes the handler based on the event
    context and the routes provided by the handler via the @app.route decorator


    The main handler can then insert 3 functions to configure the pipeline:
    + params_parser: Mandatory.  Callable.  Takes the request object optionally transforms it, and returns it wrapper in an Either.
    + pip_initiator:  Mandatory.  Callable.  Policy Information Point
    + factory_overrides: Optional. dict.  Overrides the routing factory token constructor.  Only supports S3 overrides.  For an
                                          s3 override provide a dict in the form of {'s3': callable_function}
    + handler_guard_fn: A pre-processing guard fn to determine whether the handler should be invoked.  It returns an Either.  When the handler
                        shouldnt run the Either wraps an Exception.  In this case, the request is passed directly to the responder
    """

    request = pip_initiator(build_value(event,
                                        context,
                                        env,
                                        event_source_cls,
                                        factory_overrides if factory_overrides else {}).value)

    guard_outcome = handler_guard_fn(request)

    if guard_outcome.is_right():
        result = run_pipeline(request=request,
                              params_parser=params_parser)
    else:
        result = monad.Left(build_value(event=event,
                                        context=context,
                                        env=env,
                                        status_code=app_value.HttpStatusCode(guard_outcome.error().code),
                                        error=guard_outcome.error()).value)
    return responder(result)


def run_pipeline(request: monad.EitherMonad[app_value.Request],
                 params_parser: Callable):
    return request >> log_start >> params_parser >> route_invoker


def build_value(event,
                context,
                env,
                event_source_cls=None,
                factory_overrides: dict = None,
                status_code: app_value.HttpStatusCode = None,
                error=None) -> monad.EitherMonad[app_value.Request]:
    """
    Initialises the app_value.Request object to be passed to the pipeline
    """
    req = app_value.Request(event=app_events.event_factory(event, factory_overrides, event_source_cls),
                            context=context,
                            tracer=init_tracer(env=env, aws_context=context),
                            event_time=chronos.time_now(tz=chronos.tz_utc()),
                            observer=observable.Observer() if observable.Observer().is_configured else None,
                            pip=None,
                            response=None,
                            status_code=status_code,
                            error=error)
    return monad.Right(req)


def route_invoker(request):
    return request.event.request_function(request=request)


def template_from_route_fn(route_fn: Callable) -> str | tuple:
    return app_route.RouteMap().route_pattern_from_function(route_fn)


def log_start(request):
    logger.info(msg='Start Handler',
                tracer=request.tracer,
                event=event_kind_to_log_ctx(request))
    return monad.Right(request)


def event_kind_to_log_ctx(request: app_value.Request) -> str:
    return "{event_type}:{kind}".format(event_type=type(request.event).__name__, kind=request.event.kind)


def responder(request_or_error: monad.Either) -> dict:
    """
    The app_value.Request object must be returned with the following outcomes:
    + Wrapped in an Either.
    + 'response' property with the response to be sent wrapped in an Either.  The app_value.Request.value.result.value
      must be able to be serialised to JSON; using the Serialiser protocol
    + The app_value.Request can be Right, but the contained response is left.  In this case the response needs to implement an
      error() fn which returns an object serialisable to JSON.
    + Otherwise, app_value.Request.error() should be an Either-wrapping an object which responds to error() which is JSON serialisable
    + Finally, request_or_error may be a common-or-garden monad[app.AppError]
    """
    if request_or_error.is_left() and isinstance(request_or_error.error(), Exception):
        return _body_from_base_error(request_or_error.error())
    return _body_from_pipeline_response(request_or_error)


def _body_from_pipeline_response(request):
    response = {'multiValueHeaders': build_multi_headers(request.lift().event)}

    if request.is_right() and request.value.response.is_right():
        # When the processing pipeline completes successfully and the response dict is a success
        body = request.value.response.value
        response['headers'] = build_headers(request.value.response_headers, body)
        response['statusCode'] = request.value.status_code.value if request.value.status_code else 200
        response['body'] = body.serialise()
        status = 'ok'
    elif request.is_right() and request.value.response.is_left():
        # When the processing pipeline completes successfully but the response dict is a failure
        # Get the error property from the response and serialise this.
        body = request.value.response.error().error
        response['headers'] = build_headers(request.value.response_headers, body)
        response['statusCode'] = _error_status_code(request.value)
        response['body'] = body.serialise()
        status = 'fail'
    else:
        # When the processing pipeline fails, with the error in the 'error' property of the request.
        body = request.error().error.error()
        response['headers'] = build_headers(request.error().response_headers, body)
        response[
            'statusCode'] = request.error().status_code.value if request.error().status_code else DEFAULT_FAILURE_HTTP_CODE
        response['body'] = body.serialise()
        status = 'fail'

    logger.info(msg="End Handler", tracer=request.lift().tracer, status=status)

    return response


def _body_from_base_error(error: AppError):
    body = {'headers': {}, 'multiValueHeaders': {}}
    body['statusCode'] = error.code
    body['body'] = error.serialise()

    logger.error(msg="End Handler--with base Error",
                 error=error.message if hasattr(error, 'message') else str(error),
                 status='fail')
    return body


def _error_status_code(request: Request):
    if isinstance(request.response, monad.MEither) and request.response.is_left():
        return request.response.error().code
    if request.status_code:
        return request.status_code
    return DEFAULT_SUCCESS_HTTP_CODE


def build_headers(hdrs: dict | None = None, returning_serialiser: Any | None = None) -> dict:
    if not hdrs and not returning_serialiser:
        return {}
    provided_headers = hdrs if hdrs else {}
    body_content_type = {'Content-Type': returning_serialiser.content_type} if returning_serialiser else {}
    return {**provided_headers, **body_content_type}


def build_multi_headers(event: app_value.RequestEvent) -> dict:
    """
    Only attempts to set headers for 'Set-Cookie' and mostly for session state
    """
    if event.returnable_session_state() and event.web_session:
        return event.web_session.serialise_state_as_multi_header()
    return {}


def init_tracer(env: environment.EnvironmentProtocol, aws_context=None):
    aws_request_id = aws_context.aws_request_id if aws_context else None
    return span_tracer.SpanTracer(environment=env,
                                  kv={'handler_id': aws_request_id})



from typing import Dict, List, Union, Callable
from http import cookies
from pymonad.tools import curry

from metis_fn import fn

class SessionProperty:

    def __init__(self, name: str, value: Union[cookies.Morsel, str], attributes: Dict = None):
        self.name = name
        self.attributes = attributes
        self.morsel = self.coerse_value(name, value, attributes)
        self.transformer = fn.identity

    def coerse_value(self, name, val, attributes):
        if isinstance(val, cookies.Morsel):
            return val
        morsel = cookies.Morsel()
        morsel.set(name, val, val)
        self.add_attributes(morsel, attributes)
        return morsel

    def update(self, val, attributes):
        self.morsel.set(self.name, val, val)
        self.add_attributes(self.morsel, attributes)
        pass

    def add_attributes(self, morsel, attributes):
        if not attributes:
            return morsel
        for cookie_attr in attributes.items():
            morsel[cookie_attr[0]] = cookie_attr[1]
        return morsel


    def value_transformer(self, transform_fn: Callable = fn.identity):
        
        self.transformer = transform_fn
        return self

    def serialise(self) -> str:
        return self.morsel.OutputString()

    def value(self):
        return self.transformer(self.morsel.value)

    def is_name(self, search_name):
        return self.name == search_name


class WebSession():

    def __init__(self):
        self.properties = []

    def session_from_headers(self, headers: Dict):
        if not headers:
            return self

        hdrs = headers.get('cookie', None)

        if not hdrs:
            return self
        cookie = cookies.SimpleCookie()
        cookie.load(hdrs)
        self.properties = [SessionProperty(item[0], item[1]) for item in cookie.items()]
        return self

    def serialise_state_as_multi_header(self) -> Dict[str, List]:
        if not self.properties:
            return {}
        return {'Set-Cookie': [prop.serialise() for prop in self.properties]}

    def get(self, name: str, transform_fn: Callable = fn.identity) -> SessionProperty:
        if not self.properties:
            return None
        found = fn.find(self.prop_name_predicate(name), self.properties)
        return found.value_transformer(transform_fn)

    def set(self, name, value: str, attributes: Dict = None):
        prop = self.get(name)
        if prop:
            prop.update(value, attributes)
        else:
            self.properties.append(SessionProperty(name, value, attributes))
        return self

    def clear_all(self):
        self.properties = []
        return self

    @curry(3)
    def prop_name_predicate(self, name, prop):
        return prop.is_name(name)




import requests
import backoff
from typing import Dict, Tuple, Any
from metis_fn import monad

from . import http, logger, circuit

DEFAULT_MAX_RETRIES = 2

def determine_retries():
    return circuit.max_retries() or DEFAULT_MAX_RETRIES


@circuit.circuit_breaker()
@backoff.on_predicate(backoff.expo, circuit.http_retryable_monad_failure_predicate, max_tries=determine_retries(), jitter=None)
def post(endpoint,
         body,
         auth=None,
         headers={},
         encoding='json',
         circuit_state_provider=None,
         name: str = __name__,
         http_timeout: float=5.0):
    return post_invoke(endpoint=endpoint,
                       headers=headers,
                       auth=auth,
                       body=body,
                       encoding=encoding,
                       name=name,
                       http_timeout=http_timeout)

@monad.monadic_try(name="http_adapter", exception_test_fn=http.http_response_monad(__name__, http.extract_by_content_type))
@logger.with_perf_log(perf_log_type='http', name=__name__)
def post_invoke(endpoint: str, headers: Dict, auth: Tuple, body: Any, encoding: str, name: str, http_timeout: float):
    if encoding == 'json':
        return requests.post(endpoint, auth=auth, headers=headers, json=body, timeout=http_timeout)
    else:
        return requests.post(endpoint, auth=auth, headers={**headers, **encoding_to_content_type(encoding)}, data=body, timeout=http_timeout)

@circuit.circuit_breaker()
@backoff.on_predicate(backoff.expo, circuit.monad_failure_predicate, max_tries=determine_retries(), jitter=None)
def get(endpoint,
        auth=None,
        headers={},
        circuit_state_provider=None,
        name: str = __name__,
        http_timeout: float=5.0,
        exception_test_fn: callable=None,
        error_cls: Any=None):
    return get_invoke(endpoint=endpoint,
                      headers=headers,
                      auth=auth,
                      name=name,
                      http_timeout=http_timeout,
                      exception_test_fn=exception_test_fn,
                      error_cls=error_cls)


@monad.monadic_try(name="http_adapter",
                   exception_test_fn=http.http_response_monad(__name__, http.extract_by_content_type),
                   error_cls=http.HttpError)
@logger.with_perf_log(perf_log_type='http', name=__name__)
def get_invoke(endpoint: str,
               headers: Dict,
               auth: Tuple,
               name: str,
               http_timeout: float,
               exception_test_fn: callable,
               error_cls: Any):
    return requests.get(endpoint, auth=auth)


def encoding_to_content_type(encoding):
    encodings = {'urlencoded': {'Content-Type': 'application/x-www-form-urlencoded'}}
    return encodings.get(encoding, {})



from dataclasses import dataclass, field
from typing import Protocol, Any, List, Optional
from pymonad.tools import curry

from metis_fn import fn, monad

from . import subject_token, crypto
"""
Take a request and adds a Policy information Object.

The request must contain a header containing the bearer token.  This will be parsed and the result (which may be a failure) 
is available in the PIP for a PEP to interrogate.  The token is validated by the subject_token module.  This must have been
configured with the JWKS endpoint to allow for JWT signature validation.

The PIP is controlled by the apps handler.  If it is used, it must provide the following:
+ pip_config.  An instance of PipConfig with the following properties:
    + userinfo. Bool.  Whether to get userinfo or not.  The default is False.  This means pip will not cache the subject.
    + userinfo_fn.  A function which returns the result of the userinfo endpoint.  The result must be in the form of a 
                    standard OIDC claims set, and must have a 'sub' claim.

To use the PIP with the app handler, add a wrapper function which calls pip with its config and the app.request object:
The wrapper must add the generated PIP to the request.pip property.  For example:
   > request.pip = pip.pip(pip.PipConfig(), request)
   > return monad.Right(request)


"""

@dataclass
class PipConfig:
    validate_id_token: bool = True
    userinfo: bool = False
    userinfo_get_fn: callable = None
    parse_generate_id_token_fn: Optional[callable] = field(default_factory=lambda: subject_token.parse_generate_id_token)

    def execute_userinfo_get(self):
        return self.userinfo and self.userinfo_get_fn


class UserinfoStateProtocol(Protocol):
    def __init__(self, state: Any):
        ...

    def activities(self) -> List[str]:
        """
        Returns a list of str activities providing for activity information to be provided for an activity-based decision
        """
        ...


@dataclass
class Pip:
    id_token: monad.EitherMonad[crypto.IdToken] = None
    subject: monad.EitherMonad[UserinfoStateProtocol] = None

    def token_valid(self):
        return self.id_token and self.id_token.is_right()

    def subject_activities(self):
        if not self.subject or self.subject.is_left() or not self.subject.value.activities():
            return None
        return self.subject.value.activities()

@curry(2)
def pip(config: PipConfig, request):
    info = Pip(id_token=jwt_parser(config, request))

    return fn.compose_iter([userinfo(config)], info)

@curry(2)
def userinfo(config, pip):
    if not pip.token_valid():
        return pip
    if (not config.execute_userinfo_get()) and pip.token_valid():
        return pip

    pip.subject = config.userinfo_get_fn(pip.id_token)

    return pip

def jwt_parser(config, request) -> monad.Either[Any, crypto.IdToken]:
    if config.validate_id_token:
        return config.parse_generate_id_token_fn(subject_token.parse_bearer_token(request.event.headers))
    return None



import collections.abc
import sys
from metis_fn import chronos

def nested_coerse(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = nested_coerse(d.get(k, {}), v)
        else:
            if type(v) in [str, int, bool, float, list]:
                d[k] = v
            else:
                coerse_fn = getattr(sys.modules[__name__], "coerser_{}".format(type(v).__name__), None)
                if coerse_fn:
                    d[k] = coerse_fn(v)
    return d

def coerser_datetime(dt):
    return chronos.iso8601()(dt)



from typing import Tuple, Callable, Any, Protocol
from simple_memory_cache import GLOBAL_CACHE

from metis_fn import chronos, monad, singleton
from . import http_adapter, crypto, random_retry_window, logger, circuit, cache
from .tracer import Tracer

expected_envs = ['client_id',
                 'client_secret']

token_cache = GLOBAL_CACHE.MemoryCachedVar('token_cache')

BEARER_TOKEN = "BEARER_TOKEN"  # Name of bearer token in PS

_CTX = {}

def env_set_up(env):
    return all(getattr(env, var)() for var in expected_envs)


class Error(Exception):
    def __init__(self, message="", name="", ctx={}, code=500, klass="", retryable=False):
        self.code = 500 if code is None else code
        self.retryable = retryable
        self.message = message
        self.name = name
        self.ctx = ctx
        self.klass = klass
        super().__init__(self.message)

    def error(self):
        return {'error': self.message, 'code': self.code, 'step': self.name, 'ctx': self.ctx}

    def duplicate_error(self):
        return "Duplicate" in self.message


class TokenError(Error):
    pass


class TokenEnvError(Error):
    pass


"""
The token config is setup by the library caller to provide 4 arguments:

+  token_persistence_provider.  This is a caller provided class or function which is provided with a token for persisting.
   It also allows self_token to request a previously persisted token.  The token_persistence_provider must provide the following interfaces:
   + write.  Takes a key and a value and persists the token (value) with the key (key).  It must return a monad result.
   + read. Takes a key and returns an monad wrapping an object which responds to value which returns the token.
+  env.  An object which provides the following methods:
   + client_id.  Returns the client id of the service for making token requests.
   + client_secret. Returns the client secret of the service for making token requests.
   + identity_token_endpoint. Returns the endpoint for the token request (client credentials grant).
   + bearer_token.  Returns the bearer token if it exists in the env.
   + set_env_var_with_value.  Takes a key and value and writes it to the env for retieval later with the bearer_token method.
     Returns Tuple('ok', key, value)
+ window_width.  Int in seconds defining the width of the token getting window (used for randomly selecting a position to reduce
                 competing lambdas from obtaining the token at the same time).
+ expiry_threshold. Int in seconds.  The number of seconds before the token expires that it will be refreshed.
"""


class TokenConfig(singleton.Singleton):
    default_window_width = (60 * 60)  # chance of refreshing token within 1 hour band
    default_expiry_threshold = (60 * 60)  # The room to leave before the actual token expiry

    def configure(self,
                  token_persistence_provider: cache.KeyValueCachePersistenceProviderProtocol,
                  env: Any,
                  env_ready_test_fn: Callable = env_set_up,
                  circuit_state_provider: circuit.CircuitStateProviderProtocol = None,
                  window_width: int = default_window_width,
                  expiry_threshold: int = default_expiry_threshold) -> None:
        self.token_persistence_provider = token_persistence_provider
        self.env = env
        self.env_ready_test_fn = env_ready_test_fn
        self.circuit_state_provider = circuit_state_provider
        self.window_width = window_width
        self.expiry_threshold = expiry_threshold
        pass


def token(tracer: Tracer = None):
    _CTX['tracer'] = tracer
    if not TokenConfig().env_ready_test_fn(TokenConfig().env):
        return monad.Left(TokenEnvError(message="Token can not the retrieved due to a failure in env setup"))
    result = get()
    if result.is_right() and (result.value.expired() or in_token_retry_window(result.value)):
        logger.info(msg='Self Token Cache Miss',
                    ctx={'expired': result.value.expired(), 'in_window': in_token_retry_window(result.value)},
                    tracer=tracer_from_ctx())
        invalidate_cache()
        return get()
    return result


def get():
    result = cacheable_token()
    if result.is_right():
        return crypto.parse_generate_id_token(result.value)
    else:
        return result


def cacheable_token():
    return token_cache.get()


def invalidate_cache():
    token_cache.invalidate()
    pass


@token_cache.on_first_access
def get_token():
    """
    First ever access is when the token is not set in parameter store and hence the environment:
        + Check the env anyway
        + get a new token from the token service
        + write it to parameter store
        + add it to the env and return it
    All other scenarios; when in env in either expired or not expired state
        + Get from Env
        + get a new token from the token service (check whether expired)
        + write it to parameter store (only when expired)
        + add it to the env and return it (only when expired)
    """
    # returns Either((status, name, value))
    result = bearer_token_from_env() >> from_cache >> token_service >> cache

    if result.is_right():
        return monad.Right(result.value[2])
    else:
        return result


def token_service(bearer_token):
    if bearer_token and crypto.bearer_token_valid(bearer_token) and not_in_token_retry_window(bearer_token):
        if bearer_token_from_env().value is None:
            return monad.Right(('from_cache', bearer_token))
        else:
            return monad.Right(('from_env', bearer_token))

    if bearer_token:
        not_in_wind = not_in_token_retry_window(bearer_token)
        token_state = (crypto.bearer_token_valid(bearer_token), crypto.parse_generate_id_token(bearer_token).value.exp())
    else:
        not_in_wind = None
        token_state = ("No Token", "No Exp")

    logger.info(msg='Self Token Expired or not in Window',
                ctx={'valid': token_state[0],
                     'exp': token_state[1],
                     'in_window': not_in_wind},
                tracer=tracer_from_ctx())

    result = http_adapter.post(endpoint=identity_token_endpoint(),
                               auth=(client_id(), client_secret()),
                               headers={},
                               body=TokenConfig().env.client_credentials_request(),
                               encoding='urlencoded',
                               name='token_service',
                               circuit_state_provider=TokenConfig().circuit_state_provider)

    return monad.Right(('from_grant', result.value[1]['access_token'])) if result.is_right() else monad.Left(
        build_token_error(result.error()))


def from_cache(bearer_token):
    result = cache_reader(TokenConfig().token_persistence_provider)
    if result is None or result.is_left():
        return monad.Right(bearer_token)
    return monad.Right(result.value.value)


def cache(bearer_token_tuple: Tuple[str, str]) -> monad.MEither:
    """
    Cache the token.  This is a dispatcher.  It takes the configured token persistence provider (Parameter Store or Dynamo)
    and writes to the provider.
    """
    get_location, bearer_token = bearer_token_tuple
    if get_location == 'from_env':
        return monad.Right(('ok', BEARER_TOKEN, bearer_token))

    if get_location == "from_cache":
        TokenConfig().env.set_env_var_with_value(BEARER_TOKEN, bearer_token)
        return monad.Right(('ok', BEARER_TOKEN, bearer_token))

    result = cache_writer(TokenConfig().token_persistence_provider, bearer_token)
    TokenConfig().env.set_env_var_with_value(BEARER_TOKEN, bearer_token)
    return result


def cache_reader(provider: Callable):
    if not hasattr(provider, 'read'):
        return None
    return provider.read(key=BEARER_TOKEN)


def cache_writer(provider: Callable, bearer_token: str) -> monad.MEither:
    result = provider.write(BEARER_TOKEN, bearer_token)
    if result.is_right():
        return monad.Right(('ok', BEARER_TOKEN, bearer_token))
    return result


def client_id() -> str:
    return TokenConfig().env.client_id()


def client_secret() -> str:
    return TokenConfig().env.client_secret()


def identity_token_endpoint() -> str:
    return TokenConfig().env.identity_token_endpoint()


def bearer_token_from_env():
    return monad.Right(TokenConfig().env.bearer_token())


def not_in_token_retry_window(bearer_token: str) -> bool:
    id_token = crypto.parse_generate_id_token(bearer_token)
    return random_retry_window.left_of_window(width=TokenConfig().window_width,
                                              end=id_token.value.exp() - TokenConfig().expiry_threshold,
                                              at=int(chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()])))


def in_token_retry_window(id_token):
    return random_retry_window.in_window(width=TokenConfig().window_width,
                                         end=id_token.exp() - TokenConfig().expiry_threshold,
                                         at=int(chronos.time_now(tz=chronos.tz_utc(), apply=[chronos.epoch()])))


def build_token_error(result):
    return TokenError(message=result.message,
                      name="self token",
                      ctx=result.ctx,
                      code=result.code, retryable=False)



def tracer_from_ctx():
    return _CTX['tracer']



from typing import Optional, Dict, Any, List, Callable, Union, Tuple
from pymonad.tools import curry

from metis_fn import singleton, fn, monad

from . import app


class RouteMap(singleton.Singleton):
    routes = {}

    def add_route(self, pattern: Union[str, Tuple[str,str,str]], f: Callable, opts: Dict):
        self.routes[pattern] = (f, opts)
        pass

    def no_route(self, return_template=False) -> Union[Callable, Tuple[str, Callable]]:
        no_route_route = self.routes.get('no_matching_route', None)
        if not no_route_route:
            no_route_route = (self.default_no_route, None)
        if return_template:
            return 'no_matching_routes', no_route_route[0], no_route_route[1]
        return no_route_route

    def default_no_route(self, request):
        return monad.Left(request.replace('error', app.AppError(message='no matching route', code=404)))

    def get_route(self, route: Union[str, Tuple]) -> Tuple[Union[str, Tuple], Callable]:
        if isinstance(route, str):
            match = self.routes.get(route, self.no_route())
            return route, match[0], match[1]
        possible_matching_routes = self._multi_match_determination(route,
                                                                   self.event_matches(route[0], route[1], route[2]))
        if not possible_matching_routes:
            return self.no_route(True)

        # return the route_pattern, route_fn, and route_opts
        return possible_matching_routes[0][0], possible_matching_routes[0][1][0], possible_matching_routes[0][1][1],

    def _multi_match_determination(self, route, matches):
        """
        Only for APIs with the route defined as ('API', 'GET', '/resourceBase/resource/ACollection')".

        When a path has both an instance (id-based) and collection based route defined.
        E.g. @app.route(pattern=('API', 'GET', '/resourceBase/resource/ACollection')) is the same as
             @app.route(pattern=('API', 'GET', '/resourceBase/resource/{id1}'))
        when the event path is /resourceBase/resource/ACollection (ACollection also matches {id1}
        """
        if not matches or len(matches) == 1:
            return matches  # no matching routes or only 1 route
        if all(["{" in m[0][2] for m in matches]):  # Each match is a templated path, so ambiguous
            return None
        # Otherwise the first one defined is used for the path.
        return [matches[0]]



    def route_pattern_from_function(self, route_fn: Callable):
        route_item = fn.find(self.route_function_predicate(route_fn), self.routes.items())
        if route_item:
            return route_item[0]
        return None

    @curry(3)
    def route_function_predicate(self, route_fn, route):
        return route[1][0] == route_fn

    def event_matches(self, pos1, pos2, pos3) -> Dict[Tuple, str]:
        return list(fn.select(self.match_predicate(pos1, pos2, pos3), self.routes.items()))

    @curry(5)
    def match_predicate(self, pos1, pos2, pos3, route_item: Tuple[Union[str, Tuple], Callable]):
        if isinstance(route_item[0], str):
            return None
        event_type, event_qual, event_template = route_item[0]
        if event_type == pos1 and event_qual == pos2 and self.template_matches(event_template, pos3):
            return True
        return None

    def template_matches(self, template, event):
        return self.matcher(fn.rest(template.split("/")), fn.rest(event.split("/")))

    def matcher(self, template_xs: List, event_xs: List):
        template_fst, template_rst = fn.first(template_xs), fn.rest(template_xs)
        ev_fst, ev_rst = fn.first(event_xs), fn.rest(event_xs)
        if not template_fst and not ev_fst:
            return True
        if not template_fst and ev_fst:
            return False
        if template_fst and not ev_fst:
            return False
        if not self.ismatch(template_fst, ev_fst):
            return False
        return self.matcher(template_rst, ev_rst)

    def ismatch(self, template_token, ev_token):
        return ("{" in template_token and "}" in template_token) or template_token == ev_token


def route(pattern: Union[str, Tuple[str, str, str]], opts: Dict = None):
    """
    Route Mapper
    """
    def inner(fn):
        RouteMap().add_route(pattern=pattern, f=fn, opts=opts)
    return inner


def std_noop_response(request):
    return monad.Right(request.replace('response', monad.Right({})))




from dataclasses import dataclass
from functools import partial
from typing import Callable, Protocol, Union
import os
from metis_fn import monad, fn, singleton
from metis_app import aws_client_helpers, error

SecureString = "SecureString"
String = "String"


@dataclass
class ParameterState:
    state: str
    name: str
    value: str


class ParameterEnvironmentProtocol(Protocol):

    def __init__(self, root_path: str):
        ...

    def put_parameter(self,
                      value_type: Union[String, SecureString],
                      key: str,
                      value: str,
                      client: Callable,
                      tier: str = None) -> monad.Either:
        ...

    def set_parameters_in_env(self, parameters: list[dict]) -> monad.Either:
        ...

    def set_parameter_in_env(self, mutate_env: bool, parameter) -> monad.Either[
        error.ParameterStoreError, ParameterState]:
        ...


class ParameterConfiguration(singleton.Singleton):
    """
    Optionally set up configuration in the situation where keys can be passed in without a path.
    The default path is configured here.
    And where reading and writing to the os.environ space is to be overridden
    """

    def configure(self,
                  root_path: str,
                  update_test_fn: Callable | None = fn.identity,
                  parameter_env_cls: ParameterEnvironmentProtocol | None = None):
        self.root_path = root_path
        self.update_test_fn = update_test_fn
        self.parameter_env = parameter_env_cls(self.root_path) if parameter_env_cls else None
        pass

    def param_env(self):
        return None if not getattr(self, 'parameter_env', None) else self.parameter_env

    def fn_for_update_test(self):
        if (f := getattr(self, 'update_test_fn', None)):
            return f
        return fn.identity


def aws_error_test_fn(result):
    statuses = {'200': True}
    if isinstance(result, monad.Either):
        if result.is_right() and statuses.get(str(result.value['ResponseMetadata']['HTTPStatusCode']), None):
            return result
    else:
        if statuses.get(str(result['ResponseMetadata']['HTTPStatusCode']), None):
            return monad.Right(result)
    return monad.Left(error.ParameterStoreError(result.value['ResponseMetadata']['HTTPHeaders']))


class OsEnv(ParameterEnvironmentProtocol):

    def __init__(self, root_path):
        self.root_path = root_path

    @monad.monadic_try(name="put parameter",
                       exception_test_fn=aws_error_test_fn,
                       error_cls=error.ParameterStoreError)
    def put_parameter(self, value_type, tier, key, value, client):
        if ParameterConfiguration().fn_for_update_test()(key):
            return self._update_parameter(value_type, tier, key, value, client)
        return self._create_parameter(value_type, tier, key, value, client)

    def _create_parameter(self, value_type, tier, key, param, client):
        result = client.put_parameter(Name=_to_absolute_key(key),
                                      Value=param,
                                      Type=value_type,
                                      Tier=tier if tier else 'Standard',
                                      Overwrite=False)
        return result

    def _update_parameter(self, value_type, tier, key, value, client):
        result = client.put_parameter(Name=_to_absolute_key(key),
                                      Value=value,
                                      Type=value_type,
                                      Overwrite=True)
        return result

    def set_parameters_in_env(self, parameters: list[dict]) -> monad.Either:
        return monad.Right(list(map(partial(self.set_parameter_in_env, True), parameters)))

    def set_parameter_in_env(self, mutate_env: bool, parameter: dict) -> monad.Either[
        error.ParameterStoreError, ParameterState]:
        if not mutate_env:
            return monad.Right(ParameterState(state="ok",
                                              name=parameter['Name'],
                                              value=parameter['Value']))
        name = parameter['Name'].split("/")[-1]
        os.environ[name] = parameter['Value']
        return monad.Right(ParameterState(state="ok",
                                          name=name,
                                          value=parameter['Value']))


"""
set_env_from_parameter_store
----------------------------
Reads from a Path in AWS Parameter Store (SSM) and injects each parameter as an environment variable.
Provide the path to the variables.
The SSM client is expected to be available via the aws_client_helpers.aws_ctx function, or optionally, it can be passed in
"""


def set_parameter_env_from_parameter_store(path: str,
                                           client=None):
    if not ParameterConfiguration().param_env():
        return set_env_from_parameter_store(path, client)
    result = (monad.Right(ssm_client(client)) >>
              partial(get_parameters, path) >>
              ParameterConfiguration().param_env().set_parameters_in_env >>
              test_set)
    return result


def set_env_from_parameter_store(path: str, client=None):
    result = (monad.Right(ssm_client(client)) >>
              partial(get_parameters, path) >>
              set_env >>
              test_set)
    return result


def writer(key: str,
           mutate_env: bool = True,
           value_type: str = SecureString,
           tier: str = 'Standard') -> Callable:
    """
    A HOF.  Allows for the partial application of a PS write function.  Returns the write fn
    which then takes the value to write and an optional builder fn which can transform the parameter
    before setting the env.
    :param key: The PS Key
    :param mutate_env: bool; whether to update os.environ with the read written parameter
    :param value_type: String | SecureString.  Defaults to SecureString
    :return:
    """
    return partial(write, key, mutate_env, value_type, tier)


def write(key: str,
          mutate_env: bool,
          value_type: str,
          tier: str,
          value: str,
          builder: Callable = fn.identity,
          ) -> monad.Either:
    """
    Implements a cache style interface that take a key/value
    and writes it to Parameter Store
    """
    if (param_env := ParameterConfiguration().param_env()):
        put_fn = param_env.put_parameter
    else:
        put_fn = put_parameter
    result = (monad.Right(ssm_client())
              >> partial(put_fn, value_type, tier, key, value)
              >> partial(build_parameter, key, value, builder)
              >> partial(set_env_var, mutate_env))
    return result


def get_parameter_from_store(key, with_decryption: bool = True):
    return get_parameter(key, ssm_client(), with_decryption)


#
# Helpers
#

def set_env(parameters):
    return monad.Right(list(map(partial(set_env_var, True), parameters)))


def test_set(results):
    return monad.Right(results)


def set_env_var(mutate_env: bool, parameter) -> monad.Either[error.ParameterStoreError, ParameterState]:
    if not mutate_env:
        return monad.Right(ParameterState(state="ok",
                                          name=parameter['Name'],
                                          value=parameter['Value']))
    name = parameter['Name'].split("/")[-1]
    os.environ[name] = parameter['Value']
    return monad.Right(ParameterState(state="ok",
                                      name=name,
                                      value=parameter['Value']))


def get_parameters(path, client):
    return get_parameters_paged(path, client, None, [])


def get_parameters_paged(path, client, token, parameters):
    base_kwargs = {'Path': path,
                   'WithDecryption': True,
                   'Recursive': True}
    kwargs = base_kwargs if not token else {**base_kwargs, **{'NextToken': token}}
    result = aws_error_test_fn(client.get_parameters_by_path(**kwargs))
    if result.is_left():
        return result
    if not result.value.get('NextToken'):
        return monad.Right(parameters + result.value.get("Parameters"))

    return get_parameters_paged(path,
                                client,
                                result.value.get('NextToken'),
                                parameters + result.value.get("Parameters"))


@monad.monadic_try(name="get_parameter",
                   exception_test_fn=aws_error_test_fn,
                   error_cls=error.ParameterStoreError)
def get_parameter(key, client, with_decryption: bool = True):
    return client.get_parameter(Name=_to_absolute_key(key),
                                WithDecryption=with_decryption)


@monad.monadic_try(name="put parameter",
                   exception_test_fn=aws_error_test_fn,
                   error_cls=error.ParameterStoreError)
def put_parameter(value_type, tier, key, value, client):
    if ParameterConfiguration().fn_for_update_test()(_to_absolute_key(key)):
        return update_parameter(value_type, key, value, client)
    return create_parameter(value_type, key, value, client, tier)


def create_parameter(value_type, key, param, client, tier):
    result = client.put_parameter(Name=_to_absolute_key(key),
                                  Value=param,
                                  Type=value_type,
                                  Tier=tier if tier else 'Standard',
                                  Overwrite=False)
    return result


def update_parameter(value_type, key, value, client):
    result = client.put_parameter(Name=_to_absolute_key(key),
                                  Value=value,
                                  Type=value_type,
                                  Overwrite=True)
    return result


def build_parameter(key, param, builder, _response):
    return monad.Right({'Name': key, 'Value': builder(param)})


def parameter_link(key):
    return "{}{}".format(parameter_path(), key)


def parameter_path():
    return ParameterConfiguration().root_path


def _is_absolute_path(key):
    return "/" in key


def _to_absolute_key(key):
    return key if _is_absolute_path(key) else parameter_link(key)

def ssm_client(ssm_client=None):
    if ssm_client:
        return ssm_client
    return aws_client_helpers.aws_ctx().ssm
